// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"

#include "Kismet/GameplayStatics.h"
#include "Kismet/KismetSystemLibrary.h"

class FChildNodeSubset
{
public:

	union
	{
		struct
		{
			uint32 bPositiveX : 1;
			uint32 bPositiveY : 1;
			uint32 bPositiveZ : 1;
			uint32 bNegativeX : 1;
			uint32 bNegativeY : 1;
			uint32 bNegativeZ : 1;
		};

		struct
		{
			/** Only the bits for the children on the positive side of the splits. */
			uint32 PositiveChildBits : 3;

			/** Only the bits for the children on the negative side of the splits. */
			uint32 NegativeChildBits : 3;
		};

		/** All the bits corresponding to the child bits. */
		uint32 ChildBits : 6;

		/** All the bits used to store the subset. */
		uint32 AllBits;
	};

	/** Initializes the subset to be empty. */
	FChildNodeSubset()
		: AllBits(0)
	{}

	FChildNodeSubset(uint32 PositiveBits, uint32 NegativeBits)
		: AllBits(0)
	{
		PositiveChildBits = PositiveBits;
		NegativeChildBits = NegativeBits;
	}

	/** Initializes the subset to contain a single node. */
	FChildNodeSubset(uint32 ChildRef)
		: AllBits(0)
	{
		// The positive child bits correspond to the child index, and the negative to the NOT of the child index.
		PositiveChildBits = ChildRef;
		NegativeChildBits = ~ChildRef;
	}

	/** Determines whether the subset contains a specific node. */
	bool Contains(uint32 ChildRef) const {
		// This subset contains the child if it has all the bits set that are set for the subset containing only the child node.
		const FChildNodeSubset ChildSubset(ChildRef);
		return (ChildBits & ChildSubset.ChildBits) == ChildSubset.ChildBits;
	}
	//жӼǷֻһԪ
	bool IsUnique() {
		return (PositiveChildBits ^ NegativeChildBits) == 0x7;
	}
	//Ƿȫ
	bool IsAll() {
		return ChildBits == 0x3F;
	}
};

struct FLeafNode {
	FBoxSphereBounds Bound;
	bool bCanRun;
	FLeafNode(const FBoxCenterAndExtent& InBound, bool bRunState) : Bound(InBound.Center, InBound.Extent, 0), bCanRun(bRunState) {}
	FLeafNode(const FBoxSphereBounds& InBound, bool bRunState) : Bound(InBound), bCanRun(bRunState) {}
	FLeafNode(){}
};

using FBlockID = uint32;

class NAVI_API MyOctree
{
public:
	MyOctree(const FBoxSphereBounds& RootBound, float MinExtent = 50.f) : BlockIDCounter(0) {
		FBoxCenterAndExtent Root(RootBound);
		Nodes.AddDefaulted(1);
		NodeDatas.Emplace(Root);
		this->MinExtent = MinExtent;
	}
	/*
	MyOctree() : BlockIDCounter(0) {
		FBoxCenterAndExtent Root;
		Nodes.AddDefaulted(1);
		NodeDatas.Emplace(Root);
	}
	*/
private:
	//洢8ڵΪһID
	using FNodeGroupIndex = int32;
	using FNodeIndex = int32;
	float MinExtent;

	struct FNode
	{
		FNodeGroupIndex ChildIndex;
		bool bIsValid;
		//FNodeIndex NodeIndex = INDEX_NONE;
		FNode() :ChildIndex(INDEX_NONE), bIsValid(true) {}
		FNode(bool IsValid) :ChildIndex(INDEX_NONE), bIsValid(IsValid) {}
		bool IsLeaf() const
		{
			return ChildIndex == INDEX_NONE;
		}
	};
	
	struct FNodeData {
		FBoxCenterAndExtent Bound;
		TSet<FBlockID> Blocks;
		bool bCanRun = true;
		FNodeData(const FBoxCenterAndExtent& Bound) : Bound(Bound) {}
		//FNodeData(){}
	};

	TArray<FNode> Nodes;//FNodeIndex -> FNode
	TArray<FNodeIndex> Parents;//FNodeGroupIndex(child) -> FNodeIndex(parent)
	TArray<FNodeData> NodeDatas;//FNodeIndex -> FNode
	TArray<FNodeGroupIndex> FreeList;

	FORCEINLINE FNodeIndex GroupToNodeIndex(FNodeGroupIndex Index) {
		return Index * 8 + 1;
	}

	FORCEINLINE FNodeGroupIndex NodeToGroupIndex(FNodeIndex Index) {
		return (Index - 1) / 8;
	}

	FNodeGroupIndex AllocateEightNodes(const FBoxCenterAndExtent& RootBound, FNodeIndex CurNode = INDEX_NONE) {
		if (FreeList.Num() == 0) {
			Nodes.AddDefaulted(8);
			for (int i = 0; i < 8; i++) {
				NodeDatas.Emplace(GetChildBox(RootBound, i));
			}
			return Parents.Add(CurNode);
		}
		else {
			FNodeGroupIndex Result = FreeList.Pop();
			for (int i = 0; i < 8; i++) {
				NodeDatas[GroupToNodeIndex(Result) + i].Bound = GetChildBox(RootBound, i);
				NodeDatas[GroupToNodeIndex(Result) + i].bCanRun = true;
				NodeDatas[GroupToNodeIndex(Result) + i].Blocks.Reset();

				Nodes[GroupToNodeIndex(Result) + i].ChildIndex = INDEX_NONE;
				Nodes[GroupToNodeIndex(Result) + i].bIsValid = true;
			}

			return Result;
		}
	}

	bool FreeEightNodes(FNodeGroupIndex GroupIndex) {
		if (GroupIndex == INDEX_NONE) return false;

		FNodeIndex Index = GroupToNodeIndex(GroupIndex);
		
		for (int8 i = 0; i < 8; i++)
		{
			//Nodes[Index + i] = FNode(false);
			//NodeDatas[Index + i] = FNodeData();
			Nodes[Index + i].bIsValid = false;
		}
		
		Parents[GroupIndex] = INDEX_NONE;
		FreeList.Add(GroupIndex);

		return true;
	}

	TMap<FBlockID, TSet<FNodeIndex>> BlockNodes;
	uint32 BlockIDCounter;
	inline FBlockID GetNewBlockID() {
		while (BlockNodes.Contains(++BlockIDCounter)) {}
		return BlockIDCounter;
	}
	
	void AddBlockInternal(const FBoxCenterAndExtent& BlockBound, FBlockID BlockID, FNodeIndex CurNode) {
		const FBoxCenterAndExtent CurBound = NodeDatas[CurNode].Bound;
		if (Nodes[CurNode].IsLeaf() && CurBound.Extent.X >= MinExtent && !Contain(BlockBound, CurBound)) {
			//TODO ????
			FNodeGroupIndex NewGroupIndex = AllocateEightNodes(CurBound, CurNode);
			Nodes[CurNode].ChildIndex = NewGroupIndex;
			//Nodes[CurNode].ChildIndex = AllocateEightNodes(CurBound, CurNode);
			Parents[NewGroupIndex] = CurNode;
		}
		
		if (Nodes[CurNode].IsLeaf()) {
			NodeDatas[CurNode].bCanRun = false;
			NodeDatas[CurNode].Blocks.Add(BlockID);
			BlockNodes[BlockID].Add(CurNode);
		}
		else {
			FChildNodeSubset Subset = GetIntersectingChild(CurBound, BlockBound);
			for (int i = 0; i < 8; i++) {
				if (Subset.Contains(i)) {
					AddBlockInternal(BlockBound, BlockID, GroupToNodeIndex(Nodes[CurNode].ChildIndex) + i);
				}
			}
		}
	}

	//ShrinkFromTopNotUse
	void ShrinkDownNodeInternal(FNodeIndex CurNode) {
		if (Nodes[CurNode].IsLeaf()) {
			return;
		}
		else {
			FNodeIndex Childs = GroupToNodeIndex(Nodes[CurNode].ChildIndex);
			for (int8 i = 0; i < 8; i++) {
				ShrinkDownNodeInternal(Childs + i);
			}

			bool bCanShrink = true;
			bool bRunState = NodeDatas[Childs].bCanRun;
			for (int8 i = 0; i < 8; i++) {
				if (!Nodes[Childs + i].IsLeaf() || NodeDatas[Childs + i].bCanRun != bRunState) {
					bCanShrink = false;
					break;
				}
			}

			if (bCanShrink) {
				for (int8 i = 0; i < 8; i++) {
					NodeDatas[CurNode].Blocks.Append(NodeDatas[Childs + i].Blocks);
				}
				NodeDatas[CurNode].bCanRun = bRunState;
				FreeEightNodes(Nodes[CurNode].ChildIndex);
				Nodes[CurNode].ChildIndex = INDEX_NONE;
			}
		}
	}
	inline bool IsNodeValid(FNodeIndex CurNode) {
		return CurNode != INDEX_NONE && CurNode < Nodes.Num() && Nodes[CurNode].bIsValid;
	}
	//ShrinkFromBottom
	void ShrinkUpNodeInternal(FNodeIndex CurNode) {
		if (Nodes[CurNode].IsLeaf()) {
			if (CurNode != 0) ShrinkUpNodeInternal(Parents[NodeToGroupIndex(CurNode)]);
		}
		else {
			FNodeIndex Childs = GroupToNodeIndex(Nodes[CurNode].ChildIndex);
			bool bCanShrink = true;
			bool bRunState = NodeDatas[Childs].bCanRun;
			for (int8 i = 0; i < 8; i++) {
				if (!Nodes[Childs + i].IsLeaf() || NodeDatas[Childs + i].bCanRun != bRunState) {
					bCanShrink = false;
					break;
				}
			}

			if (bCanShrink) {
				for (int8 i = 0; i < 8; i++) {
					for (FBlockID Block : NodeDatas[Childs + i].Blocks) {
						BlockNodes[Block].Remove(Childs + i);
						BlockNodes[Block].Add(CurNode);
					}
					NodeDatas[CurNode].Blocks.Append(MoveTemp(NodeDatas[Childs + i].Blocks));
				}
				NodeDatas[CurNode].bCanRun = bRunState;
				FreeEightNodes(Nodes[CurNode].ChildIndex);
				Nodes[CurNode].ChildIndex = INDEX_NONE;

				if (CurNode != 0) ShrinkUpNodeInternal(Parents[NodeToGroupIndex(CurNode)]);
			}
		}
	}

	bool HitTestWithLineInternal(FVector Start, const FVector& End, FNodeIndex Node, FVector& OutResult) {
		if (Nodes[Node].IsLeaf()) {
			OutResult = Start;
			//UKismetSystemLibrary::DrawDebugBox(o, NodeDatas[Node].Bound.Center, NodeDatas[Node].Bound.Extent, FColor::Green);
			return !NodeDatas[Node].bCanRun;
		}

		const FVector Center(NodeDatas[Node].Bound.Center);
		const FVector Dir = End - Start;
		
		const FVector NormalDir = Dir.GetSafeNormal();
		//ƽʱ޷жϼ벹
		if (FMath::IsNearlyEqual(Start.X, Center.X, 0.015f)) Start.X += NormalDir.X;
		if (FMath::IsNearlyEqual(Start.Y, Center.Y, 0.015f)) Start.Y += NormalDir.Y;
		if (FMath::IsNearlyEqual(Start.Z, Center.Z, 0.015f)) Start.Z += NormalDir.Z;
		
		FVector StartVector = Start - Center;
		const FVector T = - StartVector / Dir;

		float MinT;
		uint32 MinMask;
		float MaxT;
		uint32 MaxMask;
		if (T.X < T.Y) {
			MinT = T.X;
			MinMask = 0b001;
			MaxT = T.Y;
			MaxMask = 0b010;
		}
		else {
			MinT = T.Y;
			MinMask = 0b010;
			MaxT = T.X;
			MaxMask = 0b001;
		}

		if (T.Z < MinT) {
			MinT = T.Z;
			MinMask = 0b100;
		}

		if (T.Z > MaxT) {
			MaxT = T.Z;
			MaxMask = 0b100;
		}

		//ôдΪTпInf/Nan
		//float MidT = T.X + T.Y + T.Z - MaxT - MinT;
		uint32 MidMask = 0b111 ^ MinMask ^ MaxMask;
		float MidT = T[MidMask>>1];

		FNodeIndex Child = GroupToNodeIndex(Nodes[Node].ChildIndex);
		uint32 Index = (StartVector.X > 0) | ((StartVector.Y > 0) << 1) | ((StartVector.Z > 0) << 2);

		bool HitResult = false;

		FVector CurStart = Start;
		if (MinT >= 0 && MinT <= 1) {
			HitResult = HitTestWithLineInternal(CurStart, Start + MinT * Dir, Child + Index, OutResult);
			Index ^= MinMask;
			CurStart = Start + MinT * Dir;
		}
		if (HitResult) return HitResult;

		if (MidT >= 0 && MidT <= 1) {
			HitResult = HitTestWithLineInternal(CurStart, Start + MidT * Dir, Child + Index, OutResult);
			CurStart = Start + MidT * Dir;
			Index ^= MidMask;
		}
		if (HitResult) return HitResult;
		if (MaxT >= 0 && MaxT <= 1) {
			HitResult = HitTestWithLineInternal(CurStart, Start + MaxT * Dir, Child + Index, OutResult);
			CurStart = Start + MaxT * Dir;
			Index ^= MaxMask;
		}
		if (HitResult) return HitResult;

		return HitTestWithLineInternal(CurStart, End, Child + Index, OutResult);
	}


	/* Space Test Function Begin */

	static bool Contain(const FBoxCenterAndExtent& A, const FBoxCenterAndExtent& B) {
		VectorRegister CenterA = VectorLoadAligned(&A.Center);
		VectorRegister ExtentA = VectorLoadAligned(&A.Extent);

		VectorRegister CenterB = VectorLoadAligned(&B.Center);
		VectorRegister ExtentB = VectorLoadAligned(&B.Extent);

		return VectorAnyGreaterThan(VectorAbs(VectorSubtract(CenterA, CenterB)), VectorSubtract(ExtentA, ExtentB)) == 0;
	}

	static bool IntersectUsingVec(const FBoxCenterAndExtent& A, const FBoxCenterAndExtent& B) {
		VectorRegister CenterA = VectorLoadAligned(&A.Center);
		VectorRegister ExtentA = VectorLoadAligned(&A.Extent);

		VectorRegister CenterB = VectorLoadAligned(&B.Center);
		VectorRegister ExtentB = VectorLoadAligned(&B.Extent);

		return VectorAnyGreaterThan(VectorAbs(VectorSubtract(CenterA, CenterB)), VectorAdd(ExtentA, ExtentB)) == 0;
	}

	static bool IntersectUsingIF(const FBoxSphereBounds& A, const FBoxSphereBounds& B) {
		return FBoxSphereBounds::BoxesIntersect(A, B);
	}

	static FBoxCenterAndExtent GetChildBox(const FBoxCenterAndExtent& CurNode, int32 ChildRef) {
		float ChildOffset = CurNode.Extent.X / 2.f;

		//ת
		union MaskType {
			VectorRegister v;  VectorRegisterInt i;
		} Mask;
		Mask.v = MakeVectorRegister(1u, 2u, 4u, 8u);
		VectorRegisterInt X = VectorIntLoad1(&ChildRef);
		VectorRegisterInt A = VectorIntAnd(X, Mask.i);
		Mask.i = VectorIntCompareEQ(Mask.i, A);
		//1intfloatתõĽ1ֻҪ0ɣø㷨ת
		VectorRegister ChildOffsetReg = VectorSelect(Mask.v, VectorSetFloat1(ChildOffset), VectorSetFloat1(-ChildOffset));

		FBoxCenterAndExtent LocalBounds;
		VectorRegister ZeroW = MakeVectorRegister(1.0f, 1.0f, 1.0f, 0.0f);
		VectorStoreAligned(VectorMultiply(ZeroW, VectorAdd(VectorLoadAligned(&CurNode.Center), ChildOffsetReg)), &LocalBounds.Center);
		VectorStoreAligned(VectorMultiply(ZeroW, VectorSetFloat1(ChildOffset)), &LocalBounds.Extent);

		return LocalBounds;
	}

	/*
	static FBoxCenterAndExtent GetChildBox(const FBoxCenterAndExtent& CurNode, int32 ChildRef) {
		float ChildOffset = CurNode.Extent.X / 2;

		FBoxCenterAndExtent LocalBounds;
		LocalBounds.Extent = FVector4(ChildOffset, ChildOffset, ChildOffset, 0);
		if (ChildRef & 0x1) {
			LocalBounds.Center.X = ChildOffset + CurNode.Center.X;
		}
		else {
			LocalBounds.Center.X = CurNode.Center.X - ChildOffset;
		}

		if (ChildRef & 0x2) {
			LocalBounds.Center.Y = ChildOffset + CurNode.Center.Y;
		}
		else {
			LocalBounds.Center.Y = CurNode.Center.Y - ChildOffset;
		}

		if (ChildRef & 0x4) {
			LocalBounds.Center.Z = ChildOffset + CurNode.Center.Z;
		}
		else {
			LocalBounds.Center.Z = CurNode.Center.Z - ChildOffset;
		}
		FBoxSphereBounds temp1(CurNode.Center, CurNode.Extent, 0);
		UE_LOG(LogTemp, Warning, TEXT("%s"), *temp1.ToString());
		UE_LOG(LogTemp, Warning, TEXT("%d"), ChildRef);
		FBoxSphereBounds temp2(LocalBounds.Center, LocalBounds.Extent, 0);
		UE_LOG(LogTemp, Warning, TEXT("%s"), *temp2.ToString());
		return LocalBounds;
	}
	*/

	static FChildNodeSubset GetIntersectingChild(const FBoxCenterAndExtent& RootBound, const FBoxCenterAndExtent& QueryBound) {
		FChildNodeSubset Result;

		const VectorRegister QueryBoundsCenter = VectorLoadAligned(&QueryBound.Center);
		const VectorRegister QueryBoundsExtent = VectorLoadAligned(&QueryBound.Extent);
		const VectorRegister QueryBoundsMax = VectorAdd(QueryBoundsCenter, QueryBoundsExtent);
		const VectorRegister QueryBoundsMin = VectorSubtract(QueryBoundsCenter, QueryBoundsExtent);
		const VectorRegister BoundsCenter = VectorLoadAligned(&RootBound.Center);

		Result.PositiveChildBits = VectorMaskBits(VectorCompareGT(QueryBoundsMax, BoundsCenter)) & 0x7;
		Result.NegativeChildBits = VectorMaskBits(VectorCompareLE(QueryBoundsMin, BoundsCenter)) & 0x7;
		return Result;
	}

	/* Space Test Function End */

	int GetPointInChild(const FVector& Point, FNodeIndex Node) {
		VectorRegister Center = VectorLoadAligned(&NodeDatas[Node].Bound.Center);
		VectorRegister PointVec = VectorLoadFloat3_W0(&Point);

		return VectorAnyGreaterThan(PointVec, Center);
	}

	/* Graph Begin */
	TMap<FNodeIndex, int32> LeafMap;

	void GetLeafNodesInternal(TArray<FLeafNode>& LeafNodes, FNodeIndex CurNode, bool bAllowBlocked = true) {
		if (Nodes[CurNode].IsLeaf()) {
			if (bAllowBlocked || NodeDatas[CurNode].bCanRun) {
				LeafNodes.Emplace(NodeDatas[CurNode].Bound, NodeDatas[CurNode].bCanRun);
				LeafMap.Add(CurNode, LeafNodes.Num() - 1);
			}
		}
		else {
			for (int i = 0; i < 8; i++) {
				GetLeafNodesInternal(LeafNodes, GroupToNodeIndex(Nodes[CurNode].ChildIndex) + i, bAllowBlocked);
			}
		}
	}

	FORCEINLINE float Dist(FNodeIndex NodeA, FNodeIndex NodeB) {
		return (NodeDatas[NodeA].Bound.Center - NodeDatas[NodeB].Bound.Center).Size3();
	}
	void ConnectNode(TArray<TArray<TPair<int32, float>>>& Graph, FNodeIndex NodeA, FNodeIndex NodeB, uint32 Mask) {
		//FChildNodeSubset SubsetA(0b111, (~Mask) & 0x7);
		//FChildNodeSubset SubsetB((~Mask) & 0x7, 0b111);

		if (Nodes[NodeA].IsLeaf() && Nodes[NodeB].IsLeaf()) {
			if (!NodeDatas[NodeA].bCanRun || !NodeDatas[NodeB].bCanRun) return;
			float Distance = Dist(NodeA, NodeB);
			Graph[LeafMap[NodeA]].Add(TPair<int32, float>{LeafMap[NodeB], Distance});
			Graph[LeafMap[NodeB]].Add(TPair<int32, float>{LeafMap[NodeA], Distance});
			return;
		}
		if (Nodes[NodeA].IsLeaf()) {
			if (!NodeDatas[NodeA].bCanRun) return;
			for (uint32 j = 0; j < 4; j++) {
				uint32 Node = (j & (Mask - 1)) | ((j & (~(Mask - 1))) << 1);
				ConnectNode(Graph, NodeA, GroupToNodeIndex(Nodes[NodeB].ChildIndex) + (Node), Mask);
			}
			return;
		}
		if (Nodes[NodeB].IsLeaf()) {
			if (!NodeDatas[NodeB].bCanRun) return;
			for (uint32 j = 0; j < 4; j++) {
				uint32 Node = (j & (Mask - 1)) | ((j & (~(Mask - 1))) << 1);
				ConnectNode(Graph, GroupToNodeIndex(Nodes[NodeA].ChildIndex) + (Node | Mask), NodeB, Mask);
			}
			return;
		}

		for (uint32 j = 0; j < 4; j++) {
			uint32 Node = (j & (Mask - 1)) | ((j & (~(Mask - 1))) << 1);
			ConnectNode(Graph, GroupToNodeIndex(Nodes[NodeA].ChildIndex) + (Node | Mask), GroupToNodeIndex(Nodes[NodeB].ChildIndex) + (Node), Mask);
		}
	}
	void ConnectNode(TArray<TArray<TPair<int32, float>>>& Graph, FNodeIndex Node) {
		if (Nodes[Node].IsLeaf()) return;
		FNodeIndex Child = GroupToNodeIndex(Nodes[Node].ChildIndex);
		uint32 Mask = 1;
		for (uint32 i = 0; i < 3; i++) {
			for (uint32 j = 0; j < 4; j++) {
				uint32 Code = (j & (Mask - 1)) | ((j & (~(Mask - 1))) << 1);
				ConnectNode(Graph, Code + Child, (Code | Mask) + Child, Mask);
			}
			Mask <<= 1;
		}

		for (uint32 i = 0; i < 8; i++) {
			ConnectNode(Graph, Child + i);
		}
	}

	int32 GetPointInLeavesInternal(const FVector& Point, FNodeIndex Node) {
		if (Nodes[Node].IsLeaf()) return LeafMap[Node];
		else return GetPointInLeavesInternal(Point, GroupToNodeIndex(Nodes[Node].ChildIndex) + GetPointInChild(Point, Node));
	}

	void GetIntersactingLeavesInternal(const FBoxCenterAndExtent& QueryBound, FNodeIndex Node, TArray<int32>& Leaves) {
		if (Nodes[Node].IsLeaf()) {
			if (NodeDatas[Node].bCanRun) Leaves.Add(LeafMap[Node]);
		}
		else {
			FNodeIndex Child = GroupToNodeIndex(Nodes[Node].ChildIndex);
			FChildNodeSubset Subset = GetIntersectingChild(NodeDatas[Node].Bound, QueryBound);
			for (int i = 0; i < 8; i++) {
				if (Subset.Contains(i)) GetIntersactingLeavesInternal(QueryBound, Child + i, Leaves);
			}
		}
	}
	/* Graph End */


public:
	FBlockID AddBlock(const FBoxSphereBounds& BlockBound) {
		FBlockID BlockID = GetNewBlockID();
		BlockNodes.Add(BlockID);
		AddBlockInternal(FBoxCenterAndExtent(BlockBound), BlockID, 0);
		return BlockID;
	}
	bool RemoveBlock(FBlockID Block) {
		if (!BlockNodes.Contains(Block)) return false;
		for (FNodeIndex NodeIndex : BlockNodes[Block]) {
			NodeDatas[NodeIndex].Blocks.Remove(Block);
			if (NodeDatas[NodeIndex].Blocks.Num() == 0) {
				NodeDatas[NodeIndex].bCanRun = true;
			}
		}
		
		for (FNodeIndex NodeIndex : BlockNodes[Block]) {
			if(IsNodeValid(NodeIndex)) ShrinkUpNodeInternal(NodeIndex);
		}

		BlockNodes.Remove(Block);

		return true;
	}
	
	void GenerateGraph(TArray<FLeafNode>& OutLeafNodes, TArray<TArray<TPair<int32, float>>>& Graph) {
		LeafMap.Reset();
		OutLeafNodes.Reset();
		GetLeafNodesInternal(OutLeafNodes, 0);
		Graph.Reset();
		Graph.AddDefaulted(OutLeafNodes.Num());
		ConnectNode(Graph, 0);
	}


	void GenerateComplexGraph(TArray<FLeafNode>& OutLeafNodes, TArray<TArray<TPair<int32, float>>>& Graph, float MaxDist = 500) {
		LeafMap.Reset();
		OutLeafNodes.Reset();
		GetLeafNodesInternal(OutLeafNodes, 0, false);
		Graph.Reset();
		Graph.AddDefaulted(OutLeafNodes.Num());
		
		for (int i = 0; i < OutLeafNodes.Num() - 1; i++) {
			for (int j = i + 1; j < OutLeafNodes.Num(); j++) {
				FVector HitResult;
				float Distance = FVector::Dist(OutLeafNodes[i].Bound.Origin, OutLeafNodes[j].Bound.Origin);
				if (Distance < MaxDist && !HitTestWithLine(OutLeafNodes[i].Bound.Origin, OutLeafNodes[j].Bound.Origin, HitResult)) {
					Graph[i].Emplace(j, Distance);
					Graph[j].Emplace(i, Distance);
				}
			}
		}
	}

	bool HitTestWithLine(const FVector& Start, const FVector& End, FVector& OutResult) {
		return HitTestWithLineInternal(Start, End, 0, OutResult);
	}


	//GenerateGraphԺִ
	int32 GetPointInLeaves(const FVector& Point) {
		return GetPointInLeavesInternal(Point, 0);
	}
	//GenerateGraphԺִ
	void GetIntersactingLeaves(const FBoxSphereBounds& QueryBound, TArray<int32>& Leaves) {
		GetIntersactingLeavesInternal(FBoxCenterAndExtent(QueryBound), 0, Leaves);
	}

	/*
	const FNodeData& operator[](int32 Index) const {
		return NodeDatas[Index];
	}
	*/
private:
	
};
