// Fill out your copyright notice in the Description page of Project Settings.


#include "MyOctreeActor.h"
#include "Kismet/KismetSystemLibrary.h"

#define DEBUG_ALGORITHM
//#define DEBUG_HIT

// Sets default values
AMyOctreeActor::AMyOctreeActor()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;
}

// Called when the game starts or when spawned
void AMyOctreeActor::BeginPlay()
{
	Super::BeginPlay();
	if (AllBlocks.Num() == 0) return;
	FBoxSphereBounds RootBound = GetMaxBound();
	UE_LOG(LogTemp, Warning, TEXT("%s"), *RootBound.ToString());
	Octree = MakeShared<MyOctree>(RootBound, MinExtent);
	//Octree = MyOctree(FBoxSphereBounds(FVector(0,0,0), FVector(800, 800, 800), 0));
	//LeafNodes.Empty();

	if (!bLoopTest) {
		for (auto Actor : AllBlocks) {
			FBoxSphereBounds Bound;
			Actor->GetActorBounds(true, Bound.Origin, Bound.BoxExtent);
			Blocks.Add(Octree->AddBlock(Bound));
		}
		BuildGraph();
		AddStartAndEnd(Pawn->GetActorLocation(), Targets[0]->GetActorLocation());
		if (Pawn && Targets.Num() > 0) {
			switch (Algorithm) {
			case EAlgotithm::AStar:
				RunAStar();
				break;
			case EAlgotithm::ThetaStar:
				RunThetaStar();
				break;
			case EAlgotithm::LazyThetaStar:
				RunLazyThetaStar();
				break;
			case EAlgotithm::SmoothPath:
				RunSmoothPath();
				break;
			}
		}
		RemoveStartAndEnd();
	}
}

void AMyOctreeActor::BuildGraph()
{
	if (bComplexGraph) Octree->GenerateComplexGraph(LeafNodes, Graph);
	else Octree->GenerateGraph(LeafNodes, Graph);
	LeafNodes.AddDefaulted(2);
	Graph.AddDefaulted(2);
}

void AMyOctreeActor::AddStartAndEnd(const FVector& Start, const FVector& End, float Extent)
{
	StartLeaves.Reset();
	EndLeaves.Reset();
	int32 StartIndex = LeafNodes.Num() - 2;
	int32 EndIndex = LeafNodes.Num() - 1;


	FBoxSphereBounds Bound(Start, FVector(Extent), 0);
	LeafNodes[StartIndex].bCanRun = true;
	LeafNodes[StartIndex].Bound = Bound;
	Octree->GetIntersactingLeaves(Bound, StartLeaves);

	Bound.Origin = End;
	LeafNodes[EndIndex].bCanRun = true;
	LeafNodes[EndIndex].Bound = Bound;
	Octree->GetIntersactingLeaves(Bound, EndLeaves);

	
	for (int32 Node : StartLeaves) {
		FVector HitResult;
		if (!Octree->HitTestWithLine(Start, LeafNodes[Node].Bound.Origin, HitResult)) Graph[StartIndex].Emplace(Node, FVector::Dist(Start, LeafNodes[Node].Bound.Origin));
	}

	for (int32 Node : EndLeaves) {
		FVector HitResult;
		if (!Octree->HitTestWithLine(End, LeafNodes[Node].Bound.Origin, HitResult)) Graph[Node].Emplace(EndIndex, FVector::Dist(End, LeafNodes[Node].Bound.Origin));
	}
}

void AMyOctreeActor::RemoveStartAndEnd()
{
	int32 StartIndex = LeafNodes.Num() - 2;
	Graph[StartIndex].Reset();
	int32 EndIndex = LeafNodes.Num() - 1;
	for (int32 Node : EndLeaves) {
		Graph[Node].Pop();
	}
}

void AMyOctreeActor::RunAStar()
{
	int Num = LeafNodes.Num();
	const FVector& Start = LeafNodes[Num - 2].Bound.Origin;
	const FVector& End = LeafNodes[Num - 1].Bound.Origin;

	OpenList.Reset();
	CloseSet.Reset();
	OpenList.HeapPush({ Num - 2 , Num - 2, 0, FVector::Dist(Start, End)});
	
	while (OpenList.Num() > 0) {
		FAStarNode CurNode;
		OpenList.HeapPop(CurNode, false);
		if (CloseSet.Contains(CurNode.Index)) continue;
		CloseSet.Add(CurNode.Index, { CurNode.Parent, CurNode.CurCost });

		if (CurNode.Index == Num - 1) break;
		for (TPair<int32, float>& Line : Graph[CurNode.Index]) {
			if (CloseSet.Contains(Line.Key)) continue;
			float CurCost = CurNode.CurCost + Line.Value;
			OpenList.HeapPush({ Line.Key , CurNode.Index, CurCost, CurCost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, End) });
		}
	}
}

void AMyOctreeActor::RunThetaStar()
{
#ifdef DEBUG_ALGORITHM
	int32 Count = 0;
#endif // DEBUG_ALGORITHM
	int Num = LeafNodes.Num();
	const FVector& Start = LeafNodes[Num - 2].Bound.Origin;
	const FVector& End = LeafNodes[Num - 1].Bound.Origin;

	OpenList.Reset();
	CloseSet.Reset();
	OpenList.HeapPush({ Num - 2 , Num - 2, 0, FVector::Dist(Start, End) });

	while (OpenList.Num() > 0) {
		FAStarNode CurNode;
		OpenList.HeapPop(CurNode, false);
		if (CloseSet.Contains(CurNode.Index)) continue;
		CloseSet.Add(CurNode.Index, { CurNode.Parent, CurNode.CurCost });

		if (CurNode.Index == Num - 1) break;
		for (TPair<int32, float>& Line : Graph[CurNode.Index]) {
			if (CloseSet.Contains(Line.Key)) continue;
			float CurCost = CurNode.CurCost + Line.Value;
			OpenList.HeapPush({ Line.Key , CurNode.Index, CurCost, CurCost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, End) });

			//Path2
#ifdef DEBUG_ALGORITHM
			Count++;
#endif // DEBUG_ALGORITHM
			FVector t;
			if (!Octree->HitTestWithLine(LeafNodes[CurNode.Parent].Bound.Origin, LeafNodes[Line.Key].Bound.Origin, t)) {
				CurCost = CloseSet[CurNode.Parent].Cost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, LeafNodes[CurNode.Parent].Bound.Origin);
				OpenList.HeapPush({ Line.Key , CurNode.Parent, CurCost, CurCost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, End) });
			}
		}
	}

#ifdef DEBUG_ALGORITHM
	GEngine->AddOnScreenDebugMessage(0, 5.0f, FColor::Red, FString::Printf(TEXT("射线检测次数：%d"), Count));
#endif // DEBUG_ALGORITHM
}

void AMyOctreeActor::RunLazyThetaStar()
{
#ifdef DEBUG_ALGORITHM
	int32 Count = 0;
#endif // DEBUG_ALGORITHM
	int Num = LeafNodes.Num();
	const FVector& Start = LeafNodes[Num - 2].Bound.Origin;
	const FVector& End = LeafNodes[Num - 1].Bound.Origin;

	OpenList.Reset();
	CloseSet.Reset();
	OpenList.HeapPush({ Num - 2 , Num - 2, 0, FVector::Dist(Start, End) });

	while (OpenList.Num() > 0) {
		FAStarNode CurNode;
		OpenList.HeapPop(CurNode, false);

		if (CloseSet.Contains(CurNode.Index)) continue;
		FVector HitResult;
#ifdef DEBUG_ALGORITHM
		Count++;
#endif // DEBUG_ALGORITHM
		if (CurNode.bLazyTheta && Octree->HitTestWithLine(LeafNodes[CurNode.Parent].Bound.Origin, LeafNodes[CurNode.Index].Bound.Origin, HitResult)) {
			continue;
		}
		else {
			CloseSet.Add(CurNode.Index, { CurNode.Parent, CurNode.CurCost });
		}

		if (CurNode.Index == Num - 1) break;
		for (TPair<int32, float>& Line : Graph[CurNode.Index]) {
			if (CloseSet.Contains(Line.Key)) continue;
			float CurCost = CurNode.CurCost + Line.Value;
			OpenList.HeapPush({ Line.Key , CurNode.Index, CurCost, CurCost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, End) });

			//Path2
			CurCost = CloseSet[CurNode.Parent].Cost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, LeafNodes[CurNode.Parent].Bound.Origin);
			OpenList.HeapPush({ Line.Key , CurNode.Parent, CurCost, CurCost + FVector::Dist(LeafNodes[Line.Key].Bound.Origin, End), 1 });
		}
	}

#ifdef DEBUG_ALGORITHM
	GEngine->AddOnScreenDebugMessage(0, 5.0f, FColor::Red, FString::Printf(TEXT("射线检测次数：%d"), Count));
#endif // DEBUG_ALGORITHM
}

void AMyOctreeActor::RunSmoothPath()
{
#ifdef DEBUG_ALGORITHM
	int32 Count = 0;
#endif // DEBUG_ALGORITHM
	RunAStar();
	int CurNode = LeafNodes.Num() - 1;
	if (CloseSet.Contains(CurNode)) {
		int ParentNode = CloseSet[CurNode].Parent;
		while (ParentNode != LeafNodes.Num() - 2) {
			FVector HitResult;
#ifdef DEBUG_ALGORITHM
			Count++;
#endif // DEBUG_ALGORITHM
			if (Octree->HitTestWithLine(LeafNodes[CurNode].Bound.Origin, LeafNodes[CloseSet[ParentNode].Parent].Bound.Origin, HitResult)) {
				CloseSet[CurNode].Parent = ParentNode;
				CurNode = ParentNode;
			}
			ParentNode = CloseSet[ParentNode].Parent;
		}
		CloseSet[CurNode].Parent = ParentNode;
	}

#ifdef DEBUG_ALGORITHM
	GEngine->AddOnScreenDebugMessage(0, 5.0f, FColor::Red, FString::Printf(TEXT("射线检测次数：%d"), Count));
#endif // DEBUG_ALGORITHM
}

void AMyOctreeActor::LoopTest(float DeltaTime)
{
	if (!Octree.IsValid()) return;
	Time += DeltaTime;
	if (Time > 1.f) {
		if (bAdding) {
			FBoxSphereBounds Bound;
			AllBlocks[Counter]->GetActorBounds(true, Bound.Origin, Bound.BoxExtent);
			Blocks.Add(Octree->AddBlock(Bound));
			if (Counter == AllBlocks.Num() - 1) {
				bAdding = false;
				Counter = 0;
			}
			else {
				Counter++;
			}
		}
		else {
			Octree->RemoveBlock(Blocks[Counter]);
			if (Counter == Blocks.Num() - 1) {
				bAdding = true;
				Counter = 0;
				Blocks.Empty();
			}
			else {
				Counter++;
			}
		}

		Time = 0;
		LeafNodes.Empty();
		Octree->GenerateGraph(LeafNodes, Graph);
	}
}

// Called every frame
void AMyOctreeActor::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	if (bDrawNode) {
		for (const FLeafNode& LeafNode : LeafNodes) {
			UKismetSystemLibrary::DrawDebugBox(this, LeafNode.Bound.Origin, LeafNode.Bound.BoxExtent, LeafNode.bCanRun ? FColor::Green : FColor::Red);
		}
	}

	if (bDrawLine) {
		
		for (int i = 0; i < Graph.Num(); i++) {
			for (auto& Line : Graph[i]) {
				UKismetSystemLibrary::DrawDebugLine(this, LeafNodes[i].Bound.Origin, LeafNodes[Line.Key].Bound.Origin, FColor::Blue);
			}
		}
		
	}

	if (bLoopTest) {
		LoopTest(DeltaTime);
	}
	else {
		int End = LeafNodes.Num() - 1;
		if (CloseSet.Contains(End)) {
			int32 CurNode = End;
			while (CurNode != End - 1) {
				UKismetSystemLibrary::DrawDebugLine(this, LeafNodes[CurNode].Bound.Origin, LeafNodes[CloseSet[CurNode].Parent].Bound.Origin, FColor::Blue);
#ifdef DEBUG_HIT
				FVector HitResult;
				bool T = Octree->HitTestWithLine(LeafNodes[CloseSet[CurNode].Parent].Bound.Origin, LeafNodes[CurNode].Bound.Origin, HitResult);
				if(T) UKismetSystemLibrary::DrawDebugSphere(this, HitResult, 5, 12, FColor::Blue);
#endif // DEBUG_HIT
				CurNode = CloseSet[CurNode].Parent;
			}
		}
	}

#ifdef DEBUG_HIT
	UKismetSystemLibrary::DrawDebugLine(this, Pawn->GetActorLocation(), Targets[0]->GetActorLocation(), FColor::Red);
	FVector HitResult;
	bool T = Octree->HitTestWithLine(Pawn->GetActorLocation(), Targets[0]->GetActorLocation(), HitResult);
	if (T) UKismetSystemLibrary::DrawDebugSphere(this, HitResult, 5, 12, FColor::Yellow);
#endif // DEBUG_HIT
}

