// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include "MyOctree.h"

#include "MyOctreeActor.generated.h"


template<class T>
class TPriorityQueue : private TArray<T> {
public:
	TPriorityQueue() : TArray<T>() {

	}
	TPriorityQueue(std::initializer_list<T> InitList) : TArray<T>(InitList) {
		Heapify();
	}
	explicit TPriorityQueue(const TArray& Other) : TArray<T>(Other) {
		Heapify();
	}

	void Push(T item) {
		HeapPush(item);
	}

	void Pop(T item) {
		HeapPop(item);
	}

	int32 Num() {
		return TArray<T>::Num();
	}
};


UENUM()
enum class EAlgotithm {
	AStar,
	ThetaStar,
	LazyThetaStar,
	SmoothPath
};

UCLASS()
class NAVI_API AMyOctreeActor : public AActor
{
	GENERATED_BODY()
	
public:	
	// Sets default values for this actor's properties
	AMyOctreeActor();

	UPROPERTY(EditAnywhere)
	TArray<AActor*> AllBlocks;
	UPROPERTY(EditAnywhere)
	float MinExtent;
	UPROPERTY(EditAnywhere)
	AActor* Pawn;
	UPROPERTY(EditAnywhere)
	TArray<AActor*> Targets;

	UPROPERTY(EditAnywhere)
	bool bDrawNode;
	UPROPERTY(EditAnywhere)
	bool bDrawLine;
	UPROPERTY(EditAnywhere)
	bool bLoopTest;
	UPROPERTY(EditAnywhere)
	bool bComplexGraph;

	UPROPERTY(EditAnywhere)
	EAlgotithm Algorithm;

protected:
	// Called when the game starts or when spawned
	virtual void BeginPlay() override;
	TSharedPtr<MyOctree> Octree;
	
	FORCEINLINE FVector MinVector(const FVector& A, const FVector& B) {
		float MinX = FMath::Min(A.X, B.X);
		float MinY = FMath::Min(A.Y, B.Y);
		float MinZ = FMath::Min(A.Z, B.Z);

		return FVector(MinX, MinY, MinZ);
	}
	FORCEINLINE FVector MaxVector(const FVector& A, const FVector& B) {
		float MaxX = FMath::Max(A.X, B.X);
		float MaxY = FMath::Max(A.Y, B.Y);
		float MaxZ = FMath::Max(A.Z, B.Z);

		return FVector(MaxX, MaxY, MaxZ);
	}
	FBoxSphereBounds GetMaxBound() {
		FVector BoundMin(FLT_MAX);
		FVector BoundMax(FLT_MIN);
		for (AActor* Actor : AllBlocks) {
			FBoxSphereBounds Bound;
			Actor->GetActorBounds(true, Bound.Origin, Bound.BoxExtent);
			BoundMin = MinVector(BoundMin, Bound.Origin - Bound.BoxExtent);
			BoundMax = MaxVector(BoundMax, Bound.Origin + Bound.BoxExtent);
		}
		FVector Extent = (BoundMax - BoundMin) / 2.f;
		float MaxExtent = FMath::Max3(Extent.X, Extent.Y, Extent.Z);
		Extent.X = Extent.Y = Extent.Z = MaxExtent;
		return FBoxSphereBounds((BoundMin + BoundMax) / 2.f, Extent, 0);
	}

	/* Path Finding Start */
	TArray<FLeafNode> LeafNodes;
	TArray<TArray<TPair<int32, float>>> Graph;

	TArray<int32> StartLeaves;
	TArray<int32> EndLeaves;
	//ͼStartEndλá
	void BuildGraph();
	void AddStartAndEnd(const FVector& Start, const FVector& End, float Extent = 50.f);
	void RemoveStartAndEnd();

	FBlockID StartBlockID;
	FBlockID EndBlockID;
	void BuildGraphWithStartAndEnd(const FVector& Start, const FVector& End, float Extent = 50.f);

	struct FAStarNode {
		int32 Index;
		int32 Parent;
		float CurCost;
		float TotalCost;

		bool bLazyTheta = false; //Only For LazyThetaStar
		bool operator<(const FAStarNode& Other) const {
			return TotalCost < Other.TotalCost;
		}
	};
	TArray<FAStarNode> OpenList;
	struct FPathNode {
		int32 Parent;
		float Cost;
	};
	TMap<int32, FPathNode> CloseSet;
	void RunAStar();
	void RunThetaStar();
	void RunLazyThetaStar();
	void RunSmoothPath();

	/* Path Finding End */

	/* LoopTestBegin */
	float Time = 0;
	int Counter = 0;
	TArray<FBlockID> Blocks;
	bool bAdding = true;
	void LoopTest(float DeltaTime);
	/* LoopTestEnd */
public:	
	// Called every frame
	virtual void Tick(float DeltaTime) override;


	
};
